/*******************************************************************************
connect.cpp

basic connectivity with fasttrack platforms

modify lastconn file in the working directory: 

	1 line: host
	2 line: port
	3 line: ftoperator
	4 line: password
	5 line: ftapi license file name
	6 line: service name | 0

*******************************************************************************/



#include "ftsome.h"

FTRevision appVer = {1,0,0};
unsigned long applSign = 732;

FTContext ft_ctx = NULL;
FTAPI_CONF_REC conf = {0};

//METAMARKET CONNECTION DESCRIPTOR
FTConversation_DESC_REC mm_conv_desc;

/////////////////////////////////MM OPEN-CONV CALLBACKS////////////////////////////////////////
void MMOpenConvCallback(long ConversationID, long ResultCode, FTOpenResStruct *ResStruct){
    long res = 0;
    if((mm_conv_desc.open_conv_ResultCode = ResultCode) == FTOK){
        FTTraceAppl(FTLevel_Normal, "MMOpenConvCallback: MetaMarket connection is ON!");
        FTTraceAppl(FTLevel_Normal, "MMOpenConvCallback: Market Revision: V:%d.%d.%d",   
            ResStruct->MarketRevision[0],
            ResStruct->MarketRevision[1],
            ResStruct->MarketRevision[2]);

        mm_conv_desc.ConversationID = ConversationID;
    }else{
        FTTraceAppl(FTLevel_Normal, "metamarket connection FAILED! (%d)", ResultCode);
        if(ResultCode == FTInvalidCID){
            FTTraceAppl(FTLevel_Normal, "CID was already used on platform we try to connect with another one..");
                if((res = ConnectToServer(&conf, ft_ctx, 999)) < 0){
                FTTraceAppl(FTLevel_Normal, "ConnectToServer failed [result: %d] program exiting..", res);
                //error
                FTShutdown();
                exit(-1);
            }
        }
    }
}

void MMBrokenConvCallback(long ConversationID, FTBrokenTypeEnum BrokenType){
    FTTraceAppl(FTLevel_Error, "MMBrokenConvCallback: metamarket connection broken (%d, %d)" , ConversationID, BrokenType);
}

/////////////////////////////////MAIN////////////////////////////////////////
///
int main(int argc, char *argv[]){
    long result = FTOK;

    if(LoadConfigByFile(&conf)){
        printf("PARAMETERS ERROR! exiting..\n");
        return 1;
    }

    //create an ft-context and set its properties
    ft_ctx = FTCreateContext();
    FTSetContextAttribute(ft_ctx, CTX_COMPRESSED, "ZIP");
    FTSetContextAttribute(ft_ctx, CTX_APPL_AUTHFILE, conf.license_file);

    //initialize ft-framework
    if((result = InitializeFTAPI(&conf, ft_ctx)) != FTOK){
        FTTraceAppl(FTLevel_Normal, "InitializeFTAPI failed [result: %d] program exiting..", result);
        //error
        FTShutdown();
        return 1;
    }

    conf.OpenResProc = MMOpenConvCallback;
    conf.BrokenProc = MMBrokenConvCallback;
    conf.UserType = FTUserTrader;

    //remember that connection will not be done until you call at least one FTRun() call.
    if((result = ConnectToServer(&conf, ft_ctx, 999)) < 0){
        FTTraceAppl(FTLevel_Normal, "ConnectToServer failed [result: %d] program exiting..", result);
        //error
        FTShutdown();
        return 1;
    }

/***********************************************************************************
** MAIN EVENT LOOP **
 Calling FTRun() ensures that all requests (subscriptions, transactions, queries ..) 
 will be handled by FTAPI durig program lifecycle.
***********************************************************************************/
    unsigned long times = 0; 
    do{
        if(FTRun() != FTOK){
            FTShutdown();
            return 1;
        }
        times++;
        if(!(times % 10)){
            FTTraceAppl(FTLevel_Full, "FTRun called %lu times", times);
        }
        FTSocketSelect();
    }while(1);

    //unreachable code
    return 0;
}

