/*******************************************************************************
ordersend.cpp

example of sending ADD/RWT/DEL of an order.

*******************************************************************************/

#include "ftsome.h"
#include <string>
#include <map>

FTRevision appVer = {1,0,0};
unsigned long applSign = 732;

FTContext ft_ctx = NULL;
FTAPI_CONF_REC conf = {0};

static unsigned long securityCount = 0;

//received only by request sec info (only for example purpose)
static ClassPositionRec FT_C_ORDER_Position = {0, 0, 0};


//METAMARKET CONNECTION DESCRIPTOR
FTConversation_DESC_REC mm_conv_desc;


/***************************************************************************
*  Order repo
**************************************************************************/

std::map<unsigned long, void*> txFlying_orderMap;
std::map<std::string, void*> mrk_orderMap;

long initApplStructs(){
    return 0;
}


/***************************************************************************
*  Input Thread
**************************************************************************/

static int print_order_choiche = 0;
static int sub_done = 0;

pthread_attr_t *input_thread_attr_;
pthread_t input_thread_;

static int sndneword = 0;
static char rwtorderid[FT_C_ORDER_LEN+1] = {0};
static char delorderid[FT_C_ORDER_LEN+1] = {0};
static int txordercho = 0;

static void* input_thread_run(void *arg){
    while(1){
        if(sub_done && !print_order_choiche){
            printf("\n>>>> 1) send new order\n");
            printf("\n>>>> 2) edit an order:\n");
            printf("\n>>>> 2) del an order:\n");
            for (std::map<std::string, void*>::iterator it=mrk_orderMap.begin(); it!=mrk_orderMap.end(); ++it){
                printf("id: %s\n", it->first.c_str());
            }
            print_order_choiche = 1;
            scanf("%d", &txordercho);
        }else if(txordercho){
            switch(txordercho){
            case 1:
                sndneword = 1;
                txordercho = 0;
                break;
            case 2:
                printf("\n>>>> insert orderid [RWT]\n");
                scanf("%s", rwtorderid);
                txordercho = 0;
                break;
            case 3:
                printf("\n>>>> insert orderid [DEL]\n");
                scanf("%s", delorderid);
                txordercho = 0;
                break;
            }
        }else{
            Sleep(1000);
        }
    }
    return 0;
}

int start_input_thread(){
    return pthread_create(&input_thread_, input_thread_attr_, input_thread_run, NULL);
}


////////////////////////////////APPLICATION LOGIC ON SUBSCRIBED CLASSES ////////////////////////////////////////

long handleFT_C_ORDER(long ConversationID, unsigned long ReqID, FTEventTypeEnum EventType, FTNotifySubscribeResStruct *ResStruct){
    FT_C_ORDERPtr ftobj = (FT_C_ORDERPtr)ResStruct->MarketObject;

    switch (ResStruct->EntityAction) {
        case FTEntityKIL:
            //set new class version
            UpdateVersion(ResStruct->TimeStamp[0], &FT_C_ORDER_Position);
            SaveClassPosition("FT_C_ORDER.pos", &FT_C_ORDER_Position);
            if (ResStruct->KeyID <= 0 && !ResStruct->MarketObject){
                // This means a class reset at the server level.
                // Every entity received previously have to be invalidated!
                FTTraceAppl(FTLevel_Normal, "FTEntityKIL Action - Class Reset Signal - New version: %d", (ResStruct->TimeStamp)[0]);
            }
            else{
                FTTraceAppl(FTLevel_Normal, "FTEntityKIL Action - KeyID [%d]", ResStruct->KeyID);
            }
            break;
        case FTEntityDEL:
            FTTraceAppl(FTLevel_Normal, "EntityDEL Action - KeyID [%d]", ResStruct->KeyID);
            UpdateTimeStamp(ResStruct->TimeStamp, &FT_C_ORDER_Position);
            SaveClassPosition("FT_C_ORDER.pos", &FT_C_ORDER_Position);
            break;
        case FTEntityADD:
        case FTEntityRWT:
            FTTraceAppl(FTLevel_Normal, "EntityADD/RWT Action - KeyID [%d]", ResStruct->KeyID);

            std::map<std::string, void*>::iterator it;
            std::string oid(ftobj->OrderID);
            if((it = mrk_orderMap.find(oid)) != mrk_orderMap.end()){
                //we already have this order
                if(!strcmp(((FT_C_ORDERPtr)(it->second))->UserText, ftobj->UserText)){
                    //unsolicited income
                    FTTraceAppl(FTLevel_Normal, "mrk img unsol [%s] [%s] [%d] [%s]", ftobj->OrderID, ftobj->ClientOrderID, ftobj->Status, ftobj->UserText);
                }else{
                    FTTraceAppl(FTLevel_Normal, "mrk img recv by subscr on usr tx [%s] [%s] [%d] [%s]", ftobj->OrderID, ftobj->ClientOrderID, ftobj->Status, ftobj->UserText);
                    //from a user tx
                }
                memcpy(it->second, ftobj, sizeof(FT_C_ORDERRec));
            }else{
                FT_C_ORDERPtr new_order = (FT_C_ORDERPtr)malloc(sizeof(FT_C_ORDERRec));
                FTTraceAppl(FTLevel_Normal, "mrk ord ins by subscr [%s] [%s] [%d] [%s]", ftobj->OrderID, ftobj->ClientOrderID, ftobj->Status, ftobj->UserText);
                memcpy(new_order, ftobj, sizeof(FT_C_ORDERRec));
                mrk_orderMap[oid] = (void*)new_order;
            }

            UpdateTimeStamp(ResStruct->TimeStamp, &FT_C_ORDER_Position);
            SaveClassPosition("FT_C_ORDER.pos",&FT_C_ORDER_Position);
            break;
    }

    return 0;
}


/////////////////////////////////SUBSCRIPTION CALLBACKS////////////////////////////////////////

static void NotifyClass(long ConversationID, unsigned long ReqID, FTEventTypeEnum EventType, FTNotifySubscribeResStruct *ResStruct){
    switch(EventType){
        case FTConnectionBroken:
            FTTraceAppl(FTLevel_Error, "onMMSubscribeNotify: connection broken (%d,%d)" , ConversationID, ReqID);
            break;
        case FTClassIdle:
            FTTraceAppl(FTLevel_Normal, "idle received on class %d", ResStruct->ClassID);
            sub_done = 1;
            break;
        case FTClassInstance: {
            switch(ResStruct->ClassID){
            case FT_C_ORDER_ID:
                handleFT_C_ORDER(ConversationID, ReqID, EventType, ResStruct);
                break;
            default:
                break;
            }
        }
        default:
            break;
    }
}

static void SubscribeResProc(long convID, unsigned long ReqID, long ResultCode, FTSubscribeResStruct *resStruct){
    FTTraceAppl(FTLevel_Normal, "-------- FTOpenSubscribeResProc: classID [%lu], SubscriptionID [%lu]", resStruct->ClassID, resStruct->SubscriptionID);
    FTTraceAppl(FTLevel_Normal, "-------- Result: %s", FTGetErrorString(ResultCode));

    if (ResultCode == FTOK){
        FTTraceAppl(FTLevel_Normal, "-------- Status: %s", (resStruct->SubscribeStatus==FTSubscribeStart)? "Start" : "Stop");
        FTTraceAppl(FTLevel_Normal, "-------- New Table version: %lu; To-Reset signal: %d", resStruct->ClassVersion, resStruct->ResetClass);

        if (resStruct->ResetClass){
            // In this case your request can't be served in incremental manner!
            // Any previous record received have to be considered invalidated and
            // the server will resend all the class content.
            switch(resStruct->ClassID){
                case FT_C_ORDER_ID:
                    UpdateVersion(resStruct->ClassVersion, &FT_C_ORDER_Position);
                    SaveClassPosition("FT_C_ORDER.pos", &FT_C_ORDER_Position);
                    break;
                default:
                    FTTraceAppl(FTLevel_Normal, "-------- Class id not managed - Class %d: New version [%lu]", resStruct->ClassID, resStruct->ClassVersion);

            }
            // Received data are stored into a file
            FTTraceAppl(FTLevel_Normal, "-------- Class Reset - Class %d: New version [%lu]", resStruct->ClassID, resStruct->ClassVersion);
        }
    }
}

void doSubscr(long ConversationID){
        long res = 0;

        FTTraceAppl(FTLevel_Normal, "subscribing FT_C_ORDER class..");
        res = FTStartSubscribeClassExt(ConversationID,
                                        FTAckUnrequired,
                                        FTFlowAll,
                                        FT_C_ORDER_ID,
                                        FT_C_ORDER_Position.classVersion, FT_C_ORDER_Position.TS,
                                        FTQueryAll,
                                        FT_C_ORDERIdKey,
                                        0, 0,
                                        SubscribeResProc, NotifyClass,
                                        999, 0, 0);
}


/////////////////////////////////MM OPEN-CONV CALLBACKS////////////////////////////////////////
void MMOpenConvCallback(long ConversationID, long ResultCode, FTOpenResStruct *ResStruct){
    long res = 0;
    if((mm_conv_desc.open_conv_ResultCode = ResultCode) == FTOK){
       FTTraceAppl(FTLevel_Normal, "metamarket connection is ON!");
       FTTraceAppl(FTLevel_Normal, "Market Revision: V%d.%d.%d", ResStruct->MarketRevision[0], ResStruct->MarketRevision[1], ResStruct->MarketRevision[2]);

       mm_conv_desc.ConversationID = ConversationID;

       FTTraceAppl(FTLevel_Normal, "setting FT_C_ORDER class version and timestamp for incremental subscription..");
       LoadClassPosition("FT_C_ORDER.pos", &FT_C_ORDER_Position);

       doSubscr(ConversationID);
    }else{
        FTTraceAppl(FTLevel_Normal, "metamarket connection FAILED! (%d)", ResultCode);
        if(ResultCode == FTInvalidCID){
            FTTraceAppl(FTLevel_Normal, "CID was already used on platform we try to connect with another one..");
                if((res = ConnectToServer(&conf, ft_ctx, 999)) < 0){
                FTTraceAppl(FTLevel_Normal, "ConnectToServer failed [result: %d] program exiting..", res);
                //error
                FTShutdown();
                exit(-1);
            }
        }
    }
}

void MMBrokenConvCallback(long ConversationID, FTBrokenTypeEnum BrokenType){
    FTTraceAppl(FTLevel_Error, "onMMBrokenConvCallback: connection broken (%d, %d)" , ConversationID, BrokenType);
}


/////////////////////////////////ORDER ADD // RWT////////////////////////////////////////

static void OrderTransactionResProc(long conversationID, unsigned long ReqID, FTSendTransactionResStruct *resStruct){
    FT_C_ORDERPtr fly_img = NULL;
    FT_C_ORDERPtr mrk_img = NULL;
    Dump_TX_Res_Info(conversationID, ReqID, resStruct, resStruct->EntityAction == FTEntityADD ? "ORDER-ADD" : "ORDER-RWT");

    std::map<unsigned long, void*>::iterator it;
    if((it = txFlying_orderMap.find(ReqID)) != txFlying_orderMap.end()){
        printf("deleting flying image..\n");
        free(it->second);
        txFlying_orderMap.erase(it);
    }else{
        printf("error");
    }

    if(resStruct->ResultClassID==FT_C_ORDER_ID && resStruct->ResultObject){
        mrk_img = (FT_C_ORDERPtr)resStruct->ResultObject;
        if(resStruct->EntityAction == FTEntityADD){
            std::map<std::string, void*>::iterator it1;
            std::string oid(mrk_img->OrderID);
            if((it1 = mrk_orderMap.find(oid)) != mrk_orderMap.end()){
                FTTraceAppl(FTLevel_Normal, "mrk ord already recv by subscr, I wont ins here.. [%s] [%s] [%d] [%s]", mrk_img->OrderID, mrk_img->ClientOrderID, mrk_img->Status, mrk_img->UserText);
            }else{
                FT_C_ORDERPtr new_order = (FT_C_ORDERPtr)malloc(sizeof(FT_C_ORDERRec));
                FTTraceAppl(FTLevel_Normal, "mrk ord ins on tx res [%s] [%s] [%d] [%s]", mrk_img->OrderID, mrk_img->ClientOrderID, mrk_img->Status, mrk_img->UserText);
                memcpy(new_order, mrk_img, sizeof(FT_C_ORDERRec));
                mrk_orderMap[oid] = new_order;
            }
        }
    }

    if(resStruct->TransactionStatus==FTTransConversationClosed){
        FTTraceAppl(FTLevel_Normal, "WARNING: Transaction response lost.");
    }
}


static unsigned long orderReqID = 0;
static unsigned long nextOrderReqID(){
    return ++orderReqID;
}


static int rwtid = 0;
static void nextRWT_ID(char *utext){
    sprintf(utext, "RWT%d", rwtid++);
}

static int addid = 0;
static void nextADD_ID(char *utext){
    sprintf(utext, "ADD%d", addid++);
}

long sendNewOrder(){
    long res = 0;
    FT_C_ORDERRec order;
    memset(&order, 0, sizeof(FT_C_ORDERRec));
    unsigned long req_id = 0;

    FILE *f;

    char price[CSTR_BUF_LEN];
    char qty[CSTR_BUF_LEN];
    char tag[FT_C_USER_TEXT_LEN+1];

    if(f = fopen("order.in", "r")){
        
        fscanf(f, "%s%s%s%lu%s%s%s%s%s%s%s", 
                            order.FTSecID,
                            order.OperatorID,
                            order.Currency,
                            &order.Verb,
                            price,
                            qty,
                            order.Client.ClientID,
                            order.Client.ClientAccount,
                            order.ExchangeID,
                            order.MarketID,
                            tag);

        fclose(f);
    }else{
        throw "error";
        //error
    }

    order.OrderType = FT_C_ORDER_TYPE_Limit;
    order.TimeInForce = FT_C_TIMEINFORCE_Day;
    order.Status = FT_C_ORDER_STATUS_Active;
    order.Price = atof(price);
    order.Qty = atof(qty);
    order.RelativeQty = FT_C_QTY_TYPE_Default;
    order.TriggerMechanism = FT_C_STOP_TRIGGER_MECHANISM_BestPrice;
    order.Client.Origin = FT_C_ORIGIN_Client;
    strcpy(order.BrokerID, "DMA");
    order.ComboIndicator = FT_C_OPT_COMBO_IND_Single;
    order.CareOrder = FT_C_FLAG_Yes;
    order.InternalOrderFlag = FT_C_FLAG_Yes;
    order.CounterpartStatus = FT_C_ORDER_STATUS_Active;
    order.AutomaticFlag = FT_C_QUOTE_AUTO_Automatic;
    //sprintf(order.Tag, "algo{%s}", tag);
    
    nextADD_ID(order.UserText);
    req_id = nextOrderReqID();


    {
        FTTransactionID TransactionID;
        FT_C_ORDERPtr new_order = (FT_C_ORDERPtr)malloc(sizeof(FT_C_ORDERRec));
        FTMakeTransactionIDExt (mm_conv_desc.ConversationID, FT_C_ORDER_ID, &TransactionID);
        FTTraceAppl(FTLevel_Normal, "tx-fly ord ins [%s]", order.UserText);
        memcpy(new_order, &order, sizeof(FT_C_ORDERRec));
        
        txFlying_orderMap[req_id] = new_order;

        FTTraceAppl(FTLevel_Normal, "sending order [reqid:%d]", req_id);

        res = FTSendTransactionExt(mm_conv_desc.ConversationID, 
                                    &TransactionID, 
                                    FTEntityADD, 
                                    FT_C_ORDER_ID, 
                                    1, 
                                    &order, 
                                    OrderTransactionResProc, 
                                    FTTrue, 
                                    0, 
                                    req_id);

        FTTraceAppl(FTLevel_Normal, "FTSendTransactionExt [%s]", FTGetErrorString(res));
    }
    return res;
}

long editOrder(char *orderid){
    long res = 0;
    FT_C_ORDERRec tosend;
    unsigned long req_id = 0;

    std::map<std::string, void*>::iterator it;
    std::string oid(orderid);
    if((it = mrk_orderMap.find(oid)) != mrk_orderMap.end()){
        memcpy(&tosend, it->second, sizeof(FT_C_ORDERRec));
        memset(tosend.OrigClientOrderID, 0, sizeof(tosend.OrigClientOrderID));
    }else{
        printf("error");
    }

    req_id = nextOrderReqID();

    FILE *f;
    char price[CSTR_BUF_LEN];
    char qty[CSTR_BUF_LEN];

    if(f = fopen("order.rwt", "r")){
        fscanf(f, "%s%s", price, qty);
        fclose(f);
    }else{
        throw "error";
        //error
    }

    tosend.Price = atof(price);
    tosend.Qty = atof(qty);

    {
        FTTransactionID TransactionID;
        FT_C_ORDERPtr to_store = (FT_C_ORDERPtr)malloc(sizeof(FT_C_ORDERRec));
        FTTraceAppl(FTLevel_Normal, "tx-fly ord rwt [%s]", tosend.UserText);
        memcpy(to_store, &tosend, sizeof(FT_C_ORDERRec));
 
        txFlying_orderMap[req_id] = to_store;
        
        FTMakeTransactionIDExt(mm_conv_desc.ConversationID, FT_C_ORDER_ID, &TransactionID);
        
        res = FTSendTransactionExt(mm_conv_desc.ConversationID, 
                                    &TransactionID,
                                    FTEntityRWT,
                                    FT_C_ORDER_ID,
                                    1,
                                    &tosend,
                                    OrderTransactionResProc,
                                    FTTrue, 0, req_id);

        FTTraceAppl(FTLevel_Normal, "FTSendTransactionExt [%s]", FTGetErrorString(res));
    }
    return res;
}

long delOrder(char *orderid){
    long res = 0;
    FT_C_ORDERRec tosend;
    unsigned long req_id = 0;

    std::map<std::string, void*>::iterator it;
    std::string oid(orderid);
    if((it = mrk_orderMap.find(oid)) != mrk_orderMap.end()){
        memcpy(&tosend, it->second, sizeof(FT_C_ORDERRec));
    }else{
        printf("error");
    }

    {
        FTTransactionID TransactionID;
        FT_C_ORDERPtr to_store = (FT_C_ORDERPtr)malloc(sizeof(FT_C_ORDERRec));
        FTTraceAppl(FTLevel_Normal, "tx-fly ord rwt [%s]", tosend.UserText);
        memcpy(to_store, &tosend, sizeof(FT_C_ORDERRec));

        txFlying_orderMap[req_id] = to_store;
        
        FTMakeTransactionID (mm_conv_desc.ConversationID, &TransactionID);
        
        res = FTSendTransactionExt(mm_conv_desc.ConversationID, 
                                    &TransactionID,
                                    FTEntityDEL,
                                    FT_C_ORDER_ID,
                                    1,
                                    &tosend,
                                    OrderTransactionResProc,
                                    FTTrue, 0, req_id);

        FTTraceAppl(FTLevel_Normal, "FTSendTransactionExt [%s]", FTGetErrorString(res));
    }
    return res;
}


/////////////////////////////////MAIN////////////////////////////////////////
int main(int argc, char *argv[]){
    long result = FTOK;

    if(LoadConfigByFile(&conf)){
        printf("PARAMETERS ERROR! exiting..\n");
        return 1;
    }

    initApplStructs();

    //create an ft-context and set its properties
    ft_ctx = FTCreateContext();
    FTSetContextAttribute(ft_ctx, CTX_COMPRESSED, "ZIP");
    FTSetContextAttribute(ft_ctx, CTX_APPL_AUTHFILE, conf.license_file);

    //initialize ft-framework
    if(InitializeFTAPI(&conf, ft_ctx) != FTOK){
        //error
        FTShutdown();
        return 1;
    }

    conf.OpenResProc = MMOpenConvCallback;
    conf.BrokenProc = MMBrokenConvCallback;
    conf.UserType = FTUserTrader;

    //remember that connection will not be done until you call at least one FTRun() call.
    if(ConnectToServer(&conf, ft_ctx, 999) < 0){
        //error
        FTShutdown();
        return 1;
    }

    start_input_thread();

/***********************************************************************************
** MAIN EVENT LOOP **
 Calling FTRun() ensures that all requests (subscriptions, transactions, queries ..) 
 will be handled by FTAPI durig program lifecycle.
***********************************************************************************/
    unsigned long times = 0; 
    do{
        if(FTRun() != FTOK){
            FTShutdown();
            return 1;
        }
        times++;
        if(!(times % 10)){
            FTTraceAppl(FTLevel_Full, "FTRun called %lu times", times);
        }
        FTSocketSelect();

        if(rwtorderid[0]){
            FTTraceAppl(FTLevel_Normal, "sending order RWT: [%s]", rwtorderid);
            editOrder(rwtorderid);
            memset(rwtorderid, 0, sizeof(rwtorderid));
            print_order_choiche = 0;
        }

        if(delorderid[0]){
            FTTraceAppl(FTLevel_Normal, "sending order DEL: [%s]", delorderid);
            delOrder(delorderid);
            memset(delorderid, 0, sizeof(delorderid));
            print_order_choiche = 0;
        }

        if(sndneword){
            FTTraceAppl(FTLevel_Normal, "sending order ADD: [%s]", rwtorderid);
            sendNewOrder();
            print_order_choiche = 0;
            sndneword = 0;
        }

    }while(1);

    //unreachable code
    return 0;
}


