/*******************************************************************************
query.cpp

example of fasttrack queries.

*******************************************************************************/

#include "ftsome.h"

FTRevision appVer = {1,0,0};
unsigned long applSign = 732;

FTContext ft_ctx = NULL;
FTAPI_CONF_REC conf = {0};

//METAMARKET CONNECTION DESCRIPTOR
FTConversation_DESC_REC mm_conv_desc;

static unsigned long securityCount = 0;
static int query1_done = 0;
static int query2_done = 0;
static int query3_done = 0;

int eval_query1_done(void *ud){
    return query1_done;
}

int eval_query2_done(void *ud){
    return query2_done;
}

int eval_query3_done(void *ud){
    return query3_done;
}

/////////////////////////////////QUERY CALLBACKS////////////////////////////////////////
static void MMQueryCreateProc(long ConversationID, unsigned long ReqID, FTQueryCreateResStruct *ResStruct){
    FTTraceAppl(FTLevel_Normal, "MMQueryCreateProc: QRY CREATED");
}


void MMQueryNotifyProc_FT_C_SECURITY_LIST(long ConversationID, FTQueryNotifyStruct *QueryNotifyStruct){
    if(QueryNotifyStruct->EOQ){
        long res = 0;
        FTTraceAppl(FTLevel_Normal, "MMQueryNotifyProc: query on securities finished!");
        query1_done = 1;
    }else{
        FT_C_SECURITY_LISTPtr p = (FT_C_SECURITY_LISTPtr)QueryNotifyStruct->Entity;
        FTTraceAppl(FTLevel_Normal, "nr:%d \tPR-RF:[%s] \tFTID:[%s]", securityCount, p->PrimaryRef, p->FTSecID);
    }
}

void MMQueryNotifyProc_FT_C_PRICE_EXEC_HISTORY(long ConversationID, FTQueryNotifyStruct *QueryNotifyStruct){
    if(QueryNotifyStruct->EOQ){
        long res = 0;
        FTTraceAppl(FTLevel_Normal, "MMQueryNotifyProc: query on securities finished!");
        query2_done = 1;
    }else{
        FT_C_PRICE_EXEC_HISTORYPtr p = (FT_C_PRICE_EXEC_HISTORYPtr)QueryNotifyStruct->Entity;
    }
}

void MMQueryNotifyProc_FT_C_TICKS(long ConversationID, FTQueryNotifyStruct *QueryNotifyStruct){
    if(QueryNotifyStruct->EOQ){
        long res = 0;
        FTTraceAppl(FTLevel_Normal, "MMQueryNotifyProc: query on securities finished!");
        query3_done = 1;
    }else{
        FT_C_TICKSPtr p = (FT_C_TICKSPtr)QueryNotifyStruct->Entity;
        securityCount++;
        FTTraceAppl(FTLevel_Normal, "nr:%d \tPR-RF:[%s] \tpricetick:[%f]", securityCount, p->FTSecID, p->Tick->PriceTick);
    }
}


int Query_Dictionary(){
    int res = 0;
    FT_C_DICTIONARY_QUERY_PARRec QP = {0};
    memset(&QP, 0, sizeof(QP));
    QP.ClassID = FT_C_SECURITY_LIST_ID;
    QP.OnlyValid = FT_C_FLAG_Yes;

    strcpy(QP.ExchangeID, "HIMTF");
    strcpy(QP.MarketID, "MTF");

    res = FTQueryCreate (   mm_conv_desc.ConversationID,
                            FT_C_DICTIONARY_QUERY_ID,
                            753,
                            FT_C_DICTIONARY_QUERY_PAR_ID,
                            &QP,
                            MMQueryCreateProc,
                            MMQueryNotifyProc_FT_C_SECURITY_LIST);

    FTTraceAppl(FTLevel_Normal, "query on FT_C_SECURITY_LIST is started!");
    return res;
}

int Query_Ticks(){
    int res = 0;
    FT_C_DICTIONARY_QUERY_PARRec QP = {0};
    memset(&QP, 0, sizeof(QP));
    QP.ClassID = FT_C_TICKS_ID;
    QP.OnlyValid = FT_C_FLAG_Yes;

    res = FTQueryCreate (   mm_conv_desc.ConversationID,
                            FT_C_DICTIONARY_QUERY_ID,
                            741,
                            FT_C_DICTIONARY_QUERY_PAR_ID,
                            &QP,
                            MMQueryCreateProc,
                            MMQueryNotifyProc_FT_C_TICKS);

    FTTraceAppl(FTLevel_Normal, "query on FT_C_TICKS is started!");
    return res;
}

int Query_PriceExecHisto(){
    int res = 0;
    FT_C_QUERY_INFO_EXECRec QP = {0};
    strcpy(QP.FTSecID, "[HIMTFMTFMTFXS0953215349");

    res = FTQueryCreate (   mm_conv_desc.ConversationID,
                            FT_C_PRICE_EXEC_HISTORY_ID,
                            756,
                            FT_C_QUERY_INFO_EXEC_ID,
                            &QP,
                            MMQueryCreateProc,
                            MMQueryNotifyProc_FT_C_PRICE_EXEC_HISTORY);

    FTTraceAppl(FTLevel_Normal, "query on FT_C_PRICE_EXEC_HISTORY is started!");
    return res;
}



/////////////////////////////////MM OPEN-CONV CALLBACKS////////////////////////////////////////
void MMOpenConvCallback(long ConversationID, long ResultCode, FTOpenResStruct *ResStruct){
    long res = 0;
    if((mm_conv_desc.open_conv_ResultCode = ResultCode) == FTOK){
        FTTraceAppl(FTLevel_Normal, "MMOpenConvCallback: MetaMarket connection is ON!");
        FTTraceAppl(FTLevel_Normal, "MMOpenConvCallback: Market Revision: V:%d.%d.%d",   
                                                                ResStruct->MarketRevision[0],
                                                                ResStruct->MarketRevision[1],
                                                                ResStruct->MarketRevision[2]);
        mm_conv_desc.ConversationID = ConversationID;

        Query_Dictionary();
        FTRunCycleUntilCondition(eval_query1_done, NULL);

        //other query types, enable if needed

        //Query_Ticks();
        //FTRunCycleUntilCondition(eval_query2_done, NULL);

        //Query_PriceExecHisto();
        //FTRunCycleUntilCondition(eval_query3_done, NULL);

   }else{
        FTTraceAppl(FTLevel_Normal, "metamarket connection FAILED! (%d)", ResultCode);
        if(ResultCode == FTInvalidCID){
            FTTraceAppl(FTLevel_Normal, "CID was already used on platform we try to connect with another one..");
                if((res = ConnectToServer(&conf, ft_ctx, 999)) < 0){
                FTTraceAppl(FTLevel_Normal, "ConnectToServer failed [result: %d] program exiting..", res);
                //error
                FTShutdown();
                exit(-1);
            }
        }
    }
}

void MMBrokenConvCallback(long ConversationID, FTBrokenTypeEnum BrokenType){
    FTTraceAppl(FTLevel_Error, "MMBrokenConvCallback: metamarket connection broken (%d, %d)" , ConversationID, BrokenType);
}

/////////////////////////////////MAIN////////////////////////////////////////
int main(int argc, char *argv[]){
    long result = FTOK;

    if(LoadConfigByFile(&conf)){
        printf("PARAMETERS ERROR! exiting..\n");
        return 1;
    }

    //create an ft-context and set its properties
    ft_ctx = FTCreateContext();
    FTSetContextAttribute(ft_ctx, CTX_COMPRESSED, "ZIP");
    FTSetContextAttribute(ft_ctx, CTX_APPL_AUTHFILE, conf.license_file);

    //initialize ft-framework
    if(InitializeFTAPI(&conf, ft_ctx) != FTOK){
        //error
        FTShutdown();
        return 1;
    }

    conf.OpenResProc = MMOpenConvCallback;
    conf.BrokenProc = MMBrokenConvCallback;
    conf.UserType = FTUserTrader;

    //remember that connection will not be done until you call at least one FTRun() call.
    if(ConnectToServer(&conf, ft_ctx, 999) < 0){
        //error
        FTShutdown();
        return 1;
    }

/***********************************************************************************
** MAIN EVENT LOOP **
 Calling FTRun() ensures that all requests (subscriptions, transactions, queries ..) 
 will be handled by FTAPI durig program lifecycle.
***********************************************************************************/
    unsigned long times = 0; 
    do{
        if(FTRun() != FTOK){
            FTShutdown();
            return 1;
        }
        times++;
        if(!(times % 10)){
            FTTraceAppl(FTLevel_Full, "FTRun called %lu times", times);
        }
        FTSocketSelect();
    }while(1);

    //unreachable code
    return 0;
}